@@ -1,9 +1,9 @@
 # orm/state.py
-# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 
 """Defines instrumentation of instances.
 
@@ -12,13 +12,28 @@ defines a large part of the ORM's interactivity.
 
 """
 
+from __future__ import annotations
+
+from typing import Any
+from typing import Callable
+from typing import Dict
+from typing import Generic
+from typing import Iterable
+from typing import Optional
+from typing import Set
+from typing import Tuple
+from typing import TYPE_CHECKING
+from typing import Union
 import weakref
 
 from . import base
 from . import exc as orm_exc
 from . import interfaces
+from ._typing import _O
+from ._typing import is_collection_impl
 from .base import ATTR_WAS_SET
 from .base import INIT_OK
+from .base import LoaderCallableStatus
 from .base import NEVER_SET
 from .base import NO_VALUE
 from .base import PASSIVE_NO_INITIALIZE
@@ -29,11 +44,61 @@ from .path_registry import PathRegistry
 from .. import exc as sa_exc
 from .. import inspection
 from .. import util
+from ..util.typing import Literal
+from ..util.typing import Protocol
+
+if TYPE_CHECKING:
+    from ._typing import _IdentityKeyType
+    from ._typing import _InstanceDict
+    from ._typing import _LoaderCallable
+    from .attributes import AttributeImpl
+    from .attributes import History
+    from .base import PassiveFlag
+    from .collections import _AdaptedCollectionProtocol
+    from .identity import IdentityMap
+    from .instrumentation import ClassManager
+    from .interfaces import ORMOption
+    from .mapper import Mapper
+    from .session import Session
+    from ..engine import Row
+    from ..ext.asyncio.session import async_session as _async_provider
+    from ..ext.asyncio.session import AsyncSession
+
+if TYPE_CHECKING:
+    _sessions: weakref.WeakValueDictionary[int, Session]
+else:
+    # late-populated by session.py
+    _sessions = None
+
+
+if not TYPE_CHECKING:
+    # optionally late-provided by sqlalchemy.ext.asyncio.session
+
+    _async_provider = None  # noqa
+
+
+class _InstanceDictProto(Protocol):
+    def __call__(self) -> Optional[IdentityMap]: ...
+
+
+class _InstallLoaderCallableProto(Protocol[_O]):
+    """used at result loading time to install a _LoaderCallable callable
+    upon a specific InstanceState, which will be used to populate an
+    attribute when that attribute is accessed.
+
+    Concrete examples are per-instance deferred column loaders and
+    relationship lazy loaders.
+
+    """
+
+    def __call__(
+        self, state: InstanceState[_O], dict_: _InstanceDict, row: Row[Any]
+    ) -> None: ...
 
 
 @inspection._self_inspects
-class InstanceState(interfaces.InspectionAttrInfo):
-    """tracks state information at the instance level.
+class InstanceState(interfaces.InspectionAttrInfo, Generic[_O]):
+    """Tracks state information at the instance level.
 
     The :class:`.InstanceState` is a key object used by the
     SQLAlchemy ORM in order to track the state of an object;
@@ -51,30 +116,74 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         >>> from sqlalchemy import inspect
         >>> insp = inspect(some_mapped_object)
+        >>> insp.attrs.nickname.history
+        History(added=['new nickname'], unchanged=(), deleted=['nickname'])
 
     .. seealso::
 
-        :ref:`core_inspection_toplevel`
+        :ref:`orm_mapper_inspection_instancestate`
 
     """
 
-    session_id = None
-    key = None
-    runid = None
-    load_options = util.EMPTY_SET
-    load_path = PathRegistry.root
-    insert_order = None
-    _strong_obj = None
-    modified = False
-    expired = False
-    _deleted = False
-    _load_pending = False
-    _orphaned_outside_of_session = False
-    is_instance = True
-    identity_token = None
-    _last_known_values = ()
+    __slots__ = (
+        "__dict__",
+        "__weakref__",
+        "class_",
+        "manager",
+        "obj",
+        "committed_state",
+        "expired_attributes",
+    )
 
-    callables = ()
+    manager: ClassManager[_O]
+    session_id: Optional[int] = None
+    key: Optional[_IdentityKeyType[_O]] = None
+    runid: Optional[int] = None
+    load_options: Tuple[ORMOption, ...] = ()
+    load_path: PathRegistry = PathRegistry.root
+    insert_order: Optional[int] = None
+    _strong_obj: Optional[object] = None
+    obj: weakref.ref[_O]
+
+    committed_state: Dict[str, Any]
+
+    modified: bool = False
+    """When ``True`` the object was modified."""
+    expired: bool = False
+    """When ``True`` the object is :term:`expired`.
+
+    .. seealso::
+
+        :ref:`session_expire`
+    """
+    _deleted: bool = False
+    _load_pending: bool = False
+    _orphaned_outside_of_session: bool = False
+    is_instance: bool = True
+    identity_token: object = None
+    _last_known_values: Optional[Dict[str, Any]] = None
+
+    _instance_dict: _InstanceDictProto
+    """A weak reference, or in the default case a plain callable, that
+    returns a reference to the current :class:`.IdentityMap`, if any.
+
+    """
+    if not TYPE_CHECKING:
+
+        def _instance_dict(self):
+            """default 'weak reference' for _instance_dict"""
+            return None
+
+    expired_attributes: Set[str]
+    """The set of keys which are 'expired' to be loaded by
+    the manager's deferred scalar loader, assuming no pending
+    changes.
+
+    See also the ``unmodified`` collection which is intersected
+    against this set when a refresh operation occurs.
+    """
+
+    callables: Dict[str, Callable[[InstanceState[_O], PassiveFlag], Any]]
     """A namespace where a per-state loader callable can be associated.
 
     In SQLAlchemy 1.0, this is only used for lazy loaders / deferred
@@ -86,23 +195,18 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
     """
 
-    def __init__(self, obj, manager):
+    if not TYPE_CHECKING:
+        callables = util.EMPTY_DICT
+
+    def __init__(self, obj: _O, manager: ClassManager[_O]):
         self.class_ = obj.__class__
         self.manager = manager
         self.obj = weakref.ref(obj, self._cleanup)
         self.committed_state = {}
         self.expired_attributes = set()
 
-    expired_attributes = None
-    """The set of keys which are 'expired' to be loaded by
-       the manager's deferred scalar loader, assuming no pending
-       changes.
-
-       see also the ``unmodified`` collection which is intersected
-       against this set when a refresh operation occurs."""
-
     @util.memoized_property
-    def attrs(self):
+    def attrs(self) -> util.ReadOnlyProperties[AttributeState]:
         """Return a namespace representing each attribute on
         the mapped object, including its current value
         and history.
@@ -113,12 +217,12 @@ class InstanceState(interfaces.InspectionAttrInfo):
         since the last flush.
 
         """
-        return util.ImmutableProperties(
-            dict((key, AttributeState(self, key)) for key in self.manager)
+        return util.ReadOnlyProperties(
+            {key: AttributeState(self, key) for key in self.manager}
         )
 
     @property
-    def transient(self):
+    def transient(self) -> bool:
         """Return ``True`` if the object is :term:`transient`.
 
         .. seealso::
@@ -129,10 +233,9 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return self.key is None and not self._attached
 
     @property
-    def pending(self):
+    def pending(self) -> bool:
         """Return ``True`` if the object is :term:`pending`.
 
-
         .. seealso::
 
             :ref:`session_object_states`
@@ -141,7 +244,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return self.key is None and self._attached
 
     @property
-    def deleted(self):
+    def deleted(self) -> bool:
         """Return ``True`` if the object is :term:`deleted`.
 
         An object that is in the deleted state is guaranteed to
@@ -171,7 +274,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return self.key is not None and self._attached and self._deleted
 
     @property
-    def was_deleted(self):
+    def was_deleted(self) -> bool:
         """Return True if this object is or was previously in the
         "deleted" state and has not been reverted to persistent.
 
@@ -180,9 +283,6 @@ class InstanceState(interfaces.InspectionAttrInfo):
         or via transaction commit and enters the "detached" state,
         this flag will continue to report True.
 
-        .. versionadded:: 1.1 - added a local method form of
-           :func:`.orm.util.was_deleted`.
-
         .. seealso::
 
             :attr:`.InstanceState.deleted` - refers to the "deleted" state
@@ -195,19 +295,13 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return self._deleted
 
     @property
-    def persistent(self):
+    def persistent(self) -> bool:
         """Return ``True`` if the object is :term:`persistent`.
 
         An object that is in the persistent state is guaranteed to
         be within the :attr:`.Session.identity_map` of its parent
         :class:`.Session`.
 
-        .. versionchanged:: 1.1 The :attr:`.InstanceState.persistent`
-           accessor no longer returns True for an object that was
-           "deleted" within a flush; use the :attr:`.InstanceState.deleted`
-           accessor to detect this state.   This allows the "persistent"
-           state to guarantee membership in the identity map.
-
         .. seealso::
 
             :ref:`session_object_states`
@@ -216,7 +310,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return self.key is not None and self._attached and not self._deleted
 
     @property
-    def detached(self):
+    def detached(self) -> bool:
         """Return ``True`` if the object is :term:`detached`.
 
         .. seealso::
@@ -226,15 +320,15 @@ class InstanceState(interfaces.InspectionAttrInfo):
         """
         return self.key is not None and not self._attached
 
-    @property
-    @util.dependencies("sqlalchemy.orm.session")
-    def _attached(self, sessionlib):
+    @util.non_memoized_property
+    @util.preload_module("sqlalchemy.orm.session")
+    def _attached(self) -> bool:
         return (
             self.session_id is not None
-            and self.session_id in sessionlib._sessions
+            and self.session_id in util.preloaded.orm_session._sessions
         )
 
-    def _track_last_known_value(self, key):
+    def _track_last_known_value(self, key: str) -> None:
         """Track the last known value of a particular key after expiration
         operations.
 
@@ -242,13 +336,14 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         """
 
-        if key not in self._last_known_values:
-            self._last_known_values = dict(self._last_known_values)
-            self._last_known_values[key] = NO_VALUE
+        lkv = self._last_known_values
+        if lkv is None:
+            self._last_known_values = lkv = {}
+        if key not in lkv:
+            lkv[key] = NO_VALUE
 
     @property
-    @util.dependencies("sqlalchemy.orm.session")
-    def session(self, sessionlib):
+    def session(self) -> Optional[Session]:
         """Return the owning :class:`.Session` for this instance,
         or ``None`` if none available.
 
@@ -259,17 +354,58 @@ class InstanceState(interfaces.InspectionAttrInfo):
         Only when the transaction is completed does the object become
         fully detached under normal circumstances.
 
+        .. seealso::
+
+            :attr:`_orm.InstanceState.async_session`
+
         """
-        return sessionlib._state_session(self)
+        if self.session_id:
+            try:
+                return _sessions[self.session_id]
+            except KeyError:
+                pass
+        return None
 
     @property
-    def object(self):
+    def async_session(self) -> Optional[AsyncSession]:
+        """Return the owning :class:`_asyncio.AsyncSession` for this instance,
+        or ``None`` if none available.
+
+        This attribute is only non-None when the :mod:`sqlalchemy.ext.asyncio`
+        API is in use for this ORM object. The returned
+        :class:`_asyncio.AsyncSession` object will be a proxy for the
+        :class:`_orm.Session` object that would be returned from the
+        :attr:`_orm.InstanceState.session` attribute for this
+        :class:`_orm.InstanceState`.
+
+        .. versionadded:: 1.4.18
+
+        .. seealso::
+
+            :ref:`asyncio_toplevel`
+
+        """
+        if _async_provider is None:
+            return None
+
+        sess = self.session
+        if sess is not None:
+            return _async_provider(sess)
+        else:
+            return None
+
+    @property
+    def object(self) -> Optional[_O]:
         """Return the mapped object represented by this
-        :class:`.InstanceState`."""
+        :class:`.InstanceState`.
+
+        Returns None if the object has been garbage collected
+
+        """
         return self.obj()
 
     @property
-    def identity(self):
+    def identity(self) -> Optional[Tuple[Any, ...]]:
         """Return the mapped identity of the mapped object.
         This is the primary key identity as persisted by the ORM
         which can always be passed directly to
@@ -289,7 +425,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
             return self.key[1]
 
     @property
-    def identity_key(self):
+    def identity_key(self) -> Optional[_IdentityKeyType[_O]]:
         """Return the identity key for the mapped object.
 
         This is the key used to locate the object within
@@ -298,25 +434,27 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
 
         """
-        # TODO: just change .key to .identity_key across
-        # the board ?  probably
         return self.key
 
     @util.memoized_property
-    def parents(self):
+    def parents(self) -> Dict[int, Union[Literal[False], InstanceState[Any]]]:
         return {}
 
     @util.memoized_property
-    def _pending_mutations(self):
+    def _pending_mutations(self) -> Dict[str, PendingCollection]:
         return {}
 
     @util.memoized_property
-    def mapper(self):
+    def _empty_collections(self) -> Dict[str, _AdaptedCollectionProtocol]:
+        return {}
+
+    @util.memoized_property
+    def mapper(self) -> Mapper[_O]:
         """Return the :class:`_orm.Mapper` used for this mapped object."""
         return self.manager.mapper
 
     @property
-    def has_identity(self):
+    def has_identity(self) -> bool:
         """Return ``True`` if this object has an identity key.
 
         This should always have the same value as the
@@ -326,7 +464,12 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return bool(self.key)
 
     @classmethod
-    def _detach_states(self, states, session, to_transient=False):
+    def _detach_states(
+        self,
+        states: Iterable[InstanceState[_O]],
+        session: Session,
+        to_transient: bool = False,
+    ) -> None:
         persistent_to_detached = (
             session.dispatch.persistent_to_detached or None
         )
@@ -358,17 +501,17 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
             state._strong_obj = None
 
-    def _detach(self, session=None):
+    def _detach(self, session: Optional[Session] = None) -> None:
         if session:
             InstanceState._detach_states([self], session)
         else:
             self.session_id = self._strong_obj = None
 
-    def _dispose(self):
+    def _dispose(self) -> None:
+        # used by the test suite, apparently
         self._detach()
-        del self.obj
 
-    def _cleanup(self, ref):
+    def _cleanup(self, ref: weakref.ref[_O]) -> None:
         """Weakref callback cleanup.
 
         This callable cleans out the state when it is being garbage
@@ -396,13 +539,9 @@ class InstanceState(interfaces.InspectionAttrInfo):
             # assert self not in instance_dict._modified
 
         self.session_id = self._strong_obj = None
-        del self.obj
-
-    def obj(self):
-        return None
 
     @property
-    def dict(self):
+    def dict(self) -> _InstanceDict:
         """Return the instance dict used by the object.
 
         Under normal circumstances, this is always synonymous
@@ -420,35 +559,39 @@ class InstanceState(interfaces.InspectionAttrInfo):
         else:
             return {}
 
-    def _initialize_instance(*mixed, **kwargs):
+    def _initialize_instance(*mixed: Any, **kwargs: Any) -> None:
         self, instance, args = mixed[0], mixed[1], mixed[2:]  # noqa
         manager = self.manager
 
         manager.dispatch.init(self, args, kwargs)
 
         try:
-            return manager.original_init(*mixed[1:], **kwargs)
+            manager.original_init(*mixed[1:], **kwargs)
         except:
             with util.safe_reraise():
                 manager.dispatch.init_failure(self, args, kwargs)
 
-    def get_history(self, key, passive):
+    def get_history(self, key: str, passive: PassiveFlag) -> History:
         return self.manager[key].impl.get_history(self, self.dict, passive)
 
-    def get_impl(self, key):
+    def get_impl(self, key: str) -> AttributeImpl:
         return self.manager[key].impl
 
-    def _get_pending_mutation(self, key):
+    def _get_pending_mutation(self, key: str) -> PendingCollection:
         if key not in self._pending_mutations:
             self._pending_mutations[key] = PendingCollection()
         return self._pending_mutations[key]
 
-    def __getstate__(self):
-        state_dict = {"instance": self.obj()}
+    def __getstate__(self) -> Dict[str, Any]:
+        state_dict: Dict[str, Any] = {
+            "instance": self.obj(),
+            "class_": self.class_,
+            "committed_state": self.committed_state,
+            "expired_attributes": self.expired_attributes,
+        }
         state_dict.update(
             (k, self.__dict__[k])
             for k in (
-                "committed_state",
                 "_pending_mutations",
                 "modified",
                 "expired",
@@ -469,16 +612,13 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         return state_dict
 
-    def __setstate__(self, state_dict):
+    def __setstate__(self, state_dict: Dict[str, Any]) -> None:
         inst = state_dict["instance"]
         if inst is not None:
             self.obj = weakref.ref(inst, self._cleanup)
             self.class_ = inst.__class__
         else:
-            # None being possible here generally new as of 0.7.4
-            # due to storage of state in "parents".  "class_"
-            # also new.
-            self.obj = None
+            self.obj = lambda: None  # type: ignore
             self.class_ = state_dict["class_"]
 
         self.committed_state = state_dict.get("committed_state", {})
@@ -491,15 +631,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
         if "callables" in state_dict:
             self.callables = state_dict["callables"]
 
-            try:
-                self.expired_attributes = state_dict["expired_attributes"]
-            except KeyError:
-                self.expired_attributes = set()
-                # 0.9 and earlier compat
-                for k in list(self.callables):
-                    if self.callables[k] is self:
-                        self.expired_attributes.add(k)
-                        del self.callables[k]
+            self.expired_attributes = state_dict["expired_attributes"]
         else:
             if "expired_attributes" in state_dict:
                 self.expired_attributes = state_dict["expired_attributes"]
@@ -514,59 +646,62 @@ class InstanceState(interfaces.InspectionAttrInfo):
             ]
         )
         if self.key:
-            try:
-                self.identity_token = self.key[2]
-            except IndexError:
-                # 1.1 and earlier compat before identity_token
-                assert len(self.key) == 2
-                self.key = self.key + (None,)
-                self.identity_token = None
+            self.identity_token = self.key[2]
 
         if "load_path" in state_dict:
             self.load_path = PathRegistry.deserialize(state_dict["load_path"])
 
         state_dict["manager"](self, inst, state_dict)
 
-    def _reset(self, dict_, key):
+    def _reset(self, dict_: _InstanceDict, key: str) -> None:
         """Remove the given attribute and any
         callables associated with it."""
 
         old = dict_.pop(key, None)
-        if old is not None and self.manager[key].impl.collection:
-            self.manager[key].impl._invalidate_collection(old)
+        manager_impl = self.manager[key].impl
+        if old is not None and is_collection_impl(manager_impl):
+            manager_impl._invalidate_collection(old)
         self.expired_attributes.discard(key)
         if self.callables:
             self.callables.pop(key, None)
 
-    def _copy_callables(self, from_):
+    def _copy_callables(self, from_: InstanceState[Any]) -> None:
         if "callables" in from_.__dict__:
             self.callables = dict(from_.callables)
 
     @classmethod
-    def _instance_level_callable_processor(cls, manager, fn, key):
+    def _instance_level_callable_processor(
+        cls, manager: ClassManager[_O], fn: _LoaderCallable, key: Any
+    ) -> _InstallLoaderCallableProto[_O]:
         impl = manager[key].impl
-        if impl.collection:
+        if is_collection_impl(impl):
+            fixed_impl = impl
 
-            def _set_callable(state, dict_, row):
+            def _set_callable(
+                state: InstanceState[_O], dict_: _InstanceDict, row: Row[Any]
+            ) -> None:
                 if "callables" not in state.__dict__:
                     state.callables = {}
                 old = dict_.pop(key, None)
                 if old is not None:
-                    impl._invalidate_collection(old)
+                    fixed_impl._invalidate_collection(old)
                 state.callables[key] = fn
 
         else:
 
-            def _set_callable(state, dict_, row):
+            def _set_callable(
+                state: InstanceState[_O], dict_: _InstanceDict, row: Row[Any]
+            ) -> None:
                 if "callables" not in state.__dict__:
                     state.callables = {}
                 state.callables[key] = fn
 
         return _set_callable
 
-    def _expire(self, dict_, modified_set):
+    def _expire(
+        self, dict_: _InstanceDict, modified_set: Set[InstanceState[Any]]
+    ) -> None:
         self.expired = True
-
         if self.modified:
             modified_set.discard(self)
             self.committed_state.clear()
@@ -581,14 +716,21 @@ class InstanceState(interfaces.InspectionAttrInfo):
             del self.__dict__["parents"]
 
         self.expired_attributes.update(
-            [
-                impl.key
-                for impl in self.manager._scalar_loader_impls
-                if impl.expire_missing or impl.key in dict_
-            ]
+            [impl.key for impl in self.manager._loader_impls]
         )
 
         if self.callables:
+            # the per state loader callables we can remove here are
+            # LoadDeferredColumns, which undefers a column at the instance
+            # level that is mapped with deferred, and LoadLazyAttribute,
+            # which lazy loads a relationship at the instance level that
+            # is mapped with "noload" or perhaps "immediateload".
+            # Before 1.4, only column-based
+            # attributes could be considered to be "expired", so here they
+            # were the only ones "unexpired", which means to make them deferred
+            # again.   For the moment, as of 1.4 we also apply the same
+            # treatment relationships now, that is, an instance level lazy
+            # loader is reset in the same way as a column loader.
             for k in self.expired_attributes.intersection(self.callables):
                 del self.callables[k]
 
@@ -598,7 +740,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         if self._last_known_values:
             self._last_known_values.update(
-                (k, dict_[k]) for k in self._last_known_values if k in dict_
+                {k: dict_[k] for k in self._last_known_values if k in dict_}
             )
 
         for key in self.manager._all_key_set.intersection(dict_):
@@ -606,7 +748,12 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         self.manager.dispatch.expire(self, None)
 
-    def _expire_attributes(self, dict_, attribute_names, no_loader=False):
+    def _expire_attributes(
+        self,
+        dict_: _InstanceDict,
+        attribute_names: Iterable[str],
+        no_loader: bool = False,
+    ) -> None:
         pending = self.__dict__.get("_pending_mutations", None)
 
         callables = self.callables
@@ -621,15 +768,12 @@ class InstanceState(interfaces.InspectionAttrInfo):
                 if callables and key in callables:
                     del callables[key]
             old = dict_.pop(key, NO_VALUE)
-            if impl.collection and old is not NO_VALUE:
+            if is_collection_impl(impl) and old is not NO_VALUE:
                 impl._invalidate_collection(old)
 
-            if (
-                self._last_known_values
-                and key in self._last_known_values
-                and old is not NO_VALUE
-            ):
-                self._last_known_values[key] = old
+            lkv = self._last_known_values
+            if lkv is not None and key in lkv and old is not NO_VALUE:
+                lkv[key] = old
 
             self.committed_state.pop(key, None)
             if pending:
@@ -637,7 +781,9 @@ class InstanceState(interfaces.InspectionAttrInfo):
 
         self.manager.dispatch.expire(self, attribute_names)
 
-    def _load_expired(self, state, passive):
+    def _load_expired(
+        self, state: InstanceState[_O], passive: PassiveFlag
+    ) -> LoaderCallableStatus:
         """__call__ allows the InstanceState to act as a deferred
         callable for loading expired attributes, which is also
         serializable (picklable).
@@ -648,8 +794,13 @@ class InstanceState(interfaces.InspectionAttrInfo):
             return PASSIVE_NO_RESULT
 
         toload = self.expired_attributes.intersection(self.unmodified)
+        toload = toload.difference(
+            attr
+            for attr in toload
+            if not self.manager[attr].impl.load_on_unexpire
+        )
 
-        self.manager.deferred_scalar_loader(self, toload)
+        self.manager.expired_attribute_loader(self, toload, passive)
 
         # if the loader failed, or this
         # instance state didn't have an identity,
@@ -660,12 +811,12 @@ class InstanceState(interfaces.InspectionAttrInfo):
         return ATTR_WAS_SET
 
     @property
-    def unmodified(self):
+    def unmodified(self) -> Set[str]:
         """Return the set of keys which have no uncommitted changes"""
 
         return set(self.manager).difference(self.committed_state)
 
-    def unmodified_intersection(self, keys):
+    def unmodified_intersection(self, keys: Iterable[str]) -> Set[str]:
         """Return self.unmodified.intersection(keys)."""
 
         return (
@@ -675,11 +826,11 @@ class InstanceState(interfaces.InspectionAttrInfo):
         )
 
     @property
-    def unloaded(self):
+    def unloaded(self) -> Set[str]:
         """Return the set of keys which do not have a loaded value.
 
-        This includes expired attributes and any other attribute that
-        was never populated or modified.
+        This includes expired attributes and any other attribute that was never
+        populated or modified.
 
         """
         return (
@@ -689,33 +840,36 @@ class InstanceState(interfaces.InspectionAttrInfo):
         )
 
     @property
-    def unloaded_expirable(self):
-        """Return the set of keys which do not have a loaded value.
+    @util.deprecated(
+        "2.0",
+        "The :attr:`.InstanceState.unloaded_expirable` attribute is "
+        "deprecated.  Please use :attr:`.InstanceState.unloaded`.",
+    )
+    def unloaded_expirable(self) -> Set[str]:
+        """Synonymous with :attr:`.InstanceState.unloaded`.
 
-        This includes expired attributes and any other attribute that
-        was never populated or modified.
+        This attribute was added as an implementation-specific detail at some
+        point and should be considered to be private.
 
         """
-        return self.unloaded.intersection(
-            attr
-            for attr in self.manager
-            if self.manager[attr].impl.expire_missing
-        )
+        return self.unloaded
 
     @property
-    def _unloaded_non_object(self):
+    def _unloaded_non_object(self) -> Set[str]:
         return self.unloaded.intersection(
             attr
             for attr in self.manager
             if self.manager[attr].impl.accepts_scalar_loader
         )
 
-    def _instance_dict(self):
-        return None
-
     def _modified_event(
-        self, dict_, attr, previous, collection=False, is_userland=False
-    ):
+        self,
+        dict_: _InstanceDict,
+        attr: Optional[AttributeImpl],
+        previous: Any,
+        collection: bool = False,
+        is_userland: bool = False,
+    ) -> None:
         if attr:
             if not attr.send_modified_events:
                 return
@@ -726,6 +880,8 @@ class InstanceState(interfaces.InspectionAttrInfo):
                 )
             if attr.key not in self.committed_state or is_userland:
                 if collection:
+                    if TYPE_CHECKING:
+                        assert is_collection_impl(attr)
                     if previous is NEVER_SET:
                         if attr.key in dict_:
                             previous = dict_[attr.key]
@@ -734,8 +890,9 @@ class InstanceState(interfaces.InspectionAttrInfo):
                         previous = attr.copy(previous)
                 self.committed_state[attr.key] = previous
 
-            if attr.key in self._last_known_values:
-                self._last_known_values[attr.key] = NO_VALUE
+            lkv = self._last_known_values
+            if lkv is not None and attr.key in lkv:
+                lkv[attr.key] = NO_VALUE
 
         # assert self._strong_obj is None or self.modified
 
@@ -743,7 +900,10 @@ class InstanceState(interfaces.InspectionAttrInfo):
             self.modified = True
             instance_dict = self._instance_dict()
             if instance_dict:
+                has_modified = bool(instance_dict._modified)
                 instance_dict._modified.add(self)
+            else:
+                has_modified = False
 
             # only create _strong_obj link if attached
             # to a session
@@ -752,6 +912,20 @@ class InstanceState(interfaces.InspectionAttrInfo):
             if self.session_id:
                 self._strong_obj = inst
 
+                # if identity map already had modified objects,
+                # assume autobegin already occurred, else check
+                # for autobegin
+                if not has_modified:
+                    # inline of autobegin, to ensure session transaction
+                    # snapshot is established
+                    try:
+                        session = _sessions[self.session_id]
+                    except KeyError:
+                        pass
+                    else:
+                        if session._transaction is None:
+                            session._autobegin_t()
+
             if inst is None and attr:
                 raise orm_exc.ObjectDereferencedError(
                     "Can't emit change event for attribute '%s' - "
@@ -760,7 +934,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
                     % (self.manager[attr.key], base.state_class_str(self))
                 )
 
-    def _commit(self, dict_, keys):
+    def _commit(self, dict_: _InstanceDict, keys: Iterable[str]) -> None:
         """Commit attributes.
 
         This is used by a partial-attribute load operation to mark committed
@@ -789,7 +963,9 @@ class InstanceState(interfaces.InspectionAttrInfo):
             ):
                 del self.callables[key]
 
-    def _commit_all(self, dict_, instance_dict=None):
+    def _commit_all(
+        self, dict_: _InstanceDict, instance_dict: Optional[IdentityMap] = None
+    ) -> None:
         """commit all attributes unconditionally.
 
         This is used after a flush() or a full load/refresh
@@ -808,7 +984,11 @@ class InstanceState(interfaces.InspectionAttrInfo):
         self._commit_all_states([(self, dict_)], instance_dict)
 
     @classmethod
-    def _commit_all_states(self, iter_, instance_dict=None):
+    def _commit_all_states(
+        self,
+        iter_: Iterable[Tuple[InstanceState[Any], _InstanceDict]],
+        instance_dict: Optional[IdentityMap] = None,
+    ) -> None:
         """Mass / highly inlined version of commit_all()."""
 
         for state, dict_ in iter_:
@@ -828,7 +1008,7 @@ class InstanceState(interfaces.InspectionAttrInfo):
             state._strong_obj = None
 
 
-class AttributeState(object):
+class AttributeState:
     """Provide an inspection interface corresponding
     to a particular attribute on a particular mapped object.
 
@@ -843,12 +1023,17 @@ class AttributeState(object):
 
     """
 
-    def __init__(self, state, key):
+    __slots__ = ("state", "key")
+
+    state: InstanceState[Any]
+    key: str
+
+    def __init__(self, state: InstanceState[Any], key: str):
         self.state = state
         self.key = key
 
     @property
-    def loaded_value(self):
+    def loaded_value(self) -> Any:
         """The current value of this attribute as loaded from the database.
 
         If the value has not been loaded, or is otherwise not present
@@ -858,7 +1043,7 @@ class AttributeState(object):
         return self.state.dict.get(self.key, NO_VALUE)
 
     @property
-    def value(self):
+    def value(self) -> Any:
         """Return the value of this attribute.
 
         This operation is equivalent to accessing the object's
@@ -871,7 +1056,7 @@ class AttributeState(object):
         )
 
     @property
-    def history(self):
+    def history(self) -> History:
         """Return the current **pre-flush** change history for
         this attribute, via the :class:`.History` interface.
 
@@ -898,7 +1083,7 @@ class AttributeState(object):
         """
         return self.state.get_history(self.key, PASSIVE_NO_INITIALIZE)
 
-    def load_history(self):
+    def load_history(self) -> History:
         """Return the current **pre-flush** change history for
         this attribute, via the :class:`.History` interface.
 
@@ -920,13 +1105,11 @@ class AttributeState(object):
 
             :func:`.attributes.get_history` - underlying function
 
-        .. versionadded:: 0.9.0
-
         """
         return self.state.get_history(self.key, PASSIVE_OFF ^ INIT_OK)
 
 
-class PendingCollection(object):
+class PendingCollection:
     """A writable placeholder for an unloaded collection.
 
     Stores items appended to and removed from a collection that has not yet
@@ -935,17 +1118,25 @@ class PendingCollection(object):
 
     """
 
-    def __init__(self):
+    __slots__ = ("deleted_items", "added_items")
+
+    deleted_items: util.IdentitySet
+    added_items: util.OrderedIdentitySet
+
+    def __init__(self) -> None:
         self.deleted_items = util.IdentitySet()
         self.added_items = util.OrderedIdentitySet()
 
-    def append(self, value):
+    def merge_with_history(self, history: History) -> History:
+        return history._merge(self.added_items, self.deleted_items)
+
+    def append(self, value: Any) -> None:
         if value in self.deleted_items:
             self.deleted_items.remove(value)
         else:
             self.added_items.add(value)
 
-    def remove(self, value):
+    def remove(self, value: Any) -> None:
         if value in self.added_items:
             self.added_items.remove(value)
         else:
